import os
import numpy as np
from itertools import chain
from functools import partial
from collections import deque
from hits import fastq

def generate_suffix(num_pieces, which_piece):
    ''' Suffix to append to piece number which_piece out of num_pieces.
    '''
    if num_pieces == 1:
        suffix = ''
    else:
        # Want a fixed-width format for aesthetic reasons. 
        digits = len(str(num_pieces - 1))
        suffix = '.{0:d}.{1:0{width}d}'.format(num_pieces,
                                               which_piece,
                                               width=digits,
                                              )
    return suffix

def piece(file_name, num_pieces, which_piece, file_format, key=None):
    ''' An iterator over the lines in piece number which_piece out of
        num_pieces in file_name, with chunks defined by file_format. 
    '''
    if file_format == 'fastq':
        find_next_chunk = _find_next_fastq_read
    elif file_format == 'sam':
        find_next_chunk = partial(_find_next_sam_chunk, key=key)

    if which_piece == -1:
        # Sentinel value indicating merged experimnt
        which_piece = 0

    this_start = _find_start(file_name,
                             num_pieces,
                             which_piece,
                             find_next_chunk,
                            )
    next_start = _find_start(file_name,
                             num_pieces,
                             which_piece + 1,
                             find_next_chunk,
                            )
    
    fh = open(file_name, 'rb')
    fh.seek(this_start)
    
    while fh.tell() < next_start:
        yield fh.readline().decode()

    fh.close()

def interleaved_piece(file_name, num_pieces, which_piece):
    ''' An iterator over the lines of fastq chunks which are equivalent to 
        which_piece mod num_pieces.
    '''
    if which_piece == -1:
        # Sentinel value indicating merged experimnt
        which_piece = 0

    for i, chunk in enumerate(fastq.get_line_groups(file_name)):
        if i % num_pieces == which_piece:
            for line in chunk:
                yield line

def compromise_piece(file_name, num_pieces, which_piece, factor):
    indices = [num_pieces * i + which_piece for i in range(factor)]
    pieces = [piece(file_name, num_pieces * factor, i, 'fastq')
              for i in indices]
    return chain(*pieces)

def _find_start(file_name, num_pieces, which_piece, find_next_chunk):
    ''' Finds the offset into file_name at which piece number which_piece out
        out of num_pieces starts, given that file_name consists of chunks which
        should not be broken up. Chunks are defined by find_next_chunk, a
        function that takes an open file and finds the start
        of the next chunk.
        Nasty gotcha - this doesn't necessarily return the same starting read
        in R1 and R2 files if the read lengths for R1 and R2 are different.
    '''
    file_size = os.path.getsize(file_name)
    if which_piece >= num_pieces:
        return file_size

    nominal_piece_size = int(np.ceil(float(file_size) / num_pieces))
    nominal_start = nominal_piece_size * which_piece
    
    with open(file_name, 'rb') as fh:
        fh.seek(nominal_start)
        next_start = find_next_chunk(fh)
    
    return next_start 

def preserve_position(func):
    ''' A decorator that takes a function that takes a filehandle and makes the
        function preserve the position of the filehandle by seeking back to
        the initial position after calling the fucntion.
    '''
    def decorated_func(fh, *args, **kwargs):
        initial_position = fh.tell()
        r = func(fh, *args, **kwargs)
        fh.seek(initial_position)
        return r

    return decorated_func

@preserve_position
def _find_next_fastq_read(fh):
    ''' Returns the position in fh of the start of the next fastq read.
        Fastq reads are sets of four lines, the first of which begins with '@'
        and the third of which begins with '+'.
    '''
    # If we are at the beginning of the file, we are at the beginning of a read.
    if fh.tell() == 0:
        return 0

    # Now we are guaranteed to not be at the beginning of the file.
    # Move to the next beginning of a line, which may be the current position.
    fh.seek(-1, os.SEEK_CUR)
    fh.readline().decode()
    
    # Find the next beginning of a set of four lines for which the first
    # starts with '@' and the third starts with '+'.
    positions = deque()
    lines = deque()
    for i in range(4):
        position = fh.tell()
        positions.append(position)
        line = fh.readline().decode()
        if not line:
            return fh.tell()
        lines.append(line)

    while not(lines[0].startswith('@') and lines[2].startswith('+')):
        position = fh.tell()
        line = fh.readline().decode()
        if not line:
            return fh.tell()
        positions.popleft()
        positions.append(position)
        lines.popleft()
        lines.append(line)
    
    return positions[0]

@preserve_position
def _find_next_sam_chunk(fh, key):
    ''' Returns the position in fh of the start of the next line that is
        transformed to a different value by key than the line before it.
        If key == None, use the line itself as the value.
    '''
    if key == None:
        key = lambda x: x

    _advance_to_next_line(fh)
    
    previous_line = _get_previous_line(fh)
    if not previous_line:
        # If there is no previous line, then the start of the next fragment
        # is the start of the reads.
        return _find_start_of_reads(fh)
    
    previous_value = key(previous_line)
    while True:
        start_of_line = fh.tell()
        current_line = fh.readline().decode()
        if not current_line:
            # Reached the end of the file.
            break
        current_value = key(current_line)
        if current_value != previous_value:
            fh.seek(start_of_line)
            break

    return fh.tell()
    
@preserve_position
def _find_start_of_reads(fh):
    ''' Find the position in fh immediately following the last SAM header line.
    '''
    fh.seek(0)
    while True:
        start_of_line = fh.tell()
        line = fh.readline().decode()
        if not line:
            start_of_reads = start_of_line
            break
        elif not line.startswith('@'):
            start_of_reads = start_of_line
            break

    return start_of_reads

def _advance_to_next_line(fh):
    ''' Advances fh to the first position that is greater than or equal to the
        current position, greater than or equal to the start of reads, and
        follows a newline.
    '''
    start_of_reads = _find_start_of_reads(fh)
    
    if fh.tell() <= start_of_reads:
        fh.seek(start_of_reads)
        return
    fh.seek(-1, os.SEEK_CUR)
    fh.readline().decode()

@preserve_position
def _get_previous_line(fh):
    ''' Given fh in a state such that the current position immediately follows
        a newline, returns the line preceding the current position
        if it exists or None if it doesn't. The start of the file
        is considered to follow a newline. The SAM header has be ignored.
    '''
    start_of_reads = _find_start_of_reads(fh)
    new_lines = 0
    # Move backwards one char at a time until hitting start_of_reads or seeing
    # two newlines.
    while fh.tell() > start_of_reads:
        fh.seek(-1, os.SEEK_CUR)
        char = fh.read(1)
        if char == '\n':
            new_lines += 1
            if new_lines == 2:
                break
        if fh.tell() == start_of_reads + 1:
            # Potentially at the beginning of the file, which is treated as if
            # it follows a newline.
            new_lines += 1
            break
        fh.seek(-1, os.SEEK_CUR)
    if new_lines == 2:
        return fh.readline().decode()
    else:
        # Reached the beginning of the reads without hitting 2 newlines, so
        # there is no previous line.
        return None



# cython: language_level=3

import numpy as np
cimport cython

cdef int SOFT_CLIPPED_typed = -2
SOFT_CLIPPED = SOFT_CLIPPED_typed
cdef int GAP_typed = -1
GAP = GAP_typed

@cython.boundscheck(False)
def generate_matrices(char* query,
                      char* target,
                      int match_bonus,
                      int mismatch_penalty,
                      int insertion_penalty,
                      int deletion_penalty,
                      bint force_query_start,
                      bint force_target_start,
                      bint force_either_start,
                      bint N_matches,
                     ):
    cdef unsigned int row, col, next_col, next_row
    cdef int match_or_mismatch, diagonal, from_left, from_above, new_score, unconstrained_start
    shape = (len(query) + 1, len(target) + 1)
    scores = np.zeros(shape, int)
    cdef long[:, ::1] scores_view = scores
    row_directions = np.zeros(shape, int)
    cdef long[:, ::1] row_directions_view = row_directions
    col_directions = np.zeros(shape, int)
    cdef long[:, ::1] col_directions_view = col_directions

    # If the alignment is constrained to include the start of the query,
    # insertion penalties need to be applied to cells in the first column.
    if force_query_start: 
        for row in range(1, len(query) + 1):
            scores_view[row, 0] = scores_view[row - 1, 0] + insertion_penalty
            row_directions_view[row, 0] = -1

    # If the alignment is constrained to include the start of the target,
    # deletion penalties need to be applied to cells in the first row.
    if force_target_start:
        for col in range(1, len(target) + 1):
            scores_view[0, col] = scores_view[0, col - 1] + deletion_penalty
            col_directions_view[0, col] = -1

    unconstrained_start = not (force_query_start or force_target_start or force_either_start)

    for row in range(1, len(query) + 1):
        for col in range(1, len(target) + 1):
            if N_matches and (query[row - 1] == b'N' or target[col - 1] == b'N'):
                match_or_mismatch = match_bonus
            elif query[row - 1] == target[col - 1]:
                match_or_mismatch = match_bonus
            else:
                match_or_mismatch = mismatch_penalty
            diagonal = scores_view[row - 1, col - 1] + match_or_mismatch

            # moving left to right is a deletion
            from_left = scores_view[row, col - 1] + deletion_penalty
            # moving top to bottom is an insertion
            from_above = scores_view[row - 1, col] + insertion_penalty

            new_score = max(diagonal, from_left, from_above)
            if unconstrained_start:
                new_score = max(0, new_score)
            scores_view[row, col] = new_score
            if new_score > max_score:
                max_score = new_score
                max_row = row
                max_col = col
            if unconstrained_start and new_score == 0:
                pass
            elif new_score == diagonal:
                col_directions_view[row, col] = -1
                row_directions_view[row, col] = -1
            elif new_score == from_left:
                col_directions_view[row, col] = -1
            elif new_score == from_above:
                row_directions_view[row, col] = -1

    matrices = {'scores': scores,
                'row_directions': row_directions,
                'col_directions': col_directions,
               }
    return matrices

def backtrack_cython(char* query,
                     char* target,
                     matrices,
                     cells_seen,
                     int end_row,
                     int end_col,
                     int force_query_start,
                     int force_target_start,
                     int force_either_start,
                    ):
    cdef int row, col, next_row, next_col, target_index, query_index
    query_mappings = np.full(len(query), SOFT_CLIPPED_typed, int)
    cdef long [:] query_mappings_view = query_mappings
    target_mappings = np.full(len(target), SOFT_CLIPPED_typed, int)
    cdef long [:] target_mappings_view = target_mappings

    cdef long [:, :] col_directions = matrices['col_directions']
    cdef long [:, :] row_directions = matrices['row_directions']
    cdef long [:, :] scores = matrices['scores']

    path = []
    insertions = set()
    deletions = set()
    mismatches = set()
    
    unconstrained_start = not(force_query_start or force_target_start or force_either_start)

    row = end_row
    col = end_col

    if row == 0 or col == 0:
        # There are no query or target bases involved in a path that ends on
        # the top or the left edge.
        reached_end = True
    else:
        reached_end = False

    while not reached_end:
        if (row, col) in cells_seen:
            return None
        cells_seen.add((row, col))

        next_col = col + col_directions[row, col]
        next_row = row + row_directions[row, col]
        if next_col == col:
            target_index = GAP_typed
            insertions.add(row - 1)
        else:
            target_index = col - 1
        if next_row == row:
            query_index = GAP_typed
            deletions.add(col - 1)
        else:
            query_index = row - 1
        
        if target_index != GAP_typed:
            target_mappings_view[target_index] = query_index
        if query_index != GAP_typed:
            query_mappings_view[query_index] = target_index
        if target_index != GAP_typed and query_index != GAP_typed and query[query_index] != b'N' and target[target_index] != b'N' and query[query_index] != target[target_index]:
            mismatches.add((query_index, target_index))

        path.append((query_index, target_index))

        row = next_row
        col = next_col

        if unconstrained_start:
            if scores[row, col] <= 0:
                reached_end = True
        elif force_query_start and force_target_start:
            if row == 0 and col == 0:
                reached_end = True
        elif force_either_start:
            if row == 0 or col == 0:
                reached_end = True
        elif force_query_start:
            if row == 0:
                reached_end = True
        elif force_target_start:
            if col == 0:
                reached_end = True

    path = path[::-1]

    alignment = {'score': scores[end_row, end_col],
                 'path': path,
                 'query_mappings': query_mappings,
                 'target_mappings': target_mappings,
                 'insertions': insertions,
                 'deletions': deletions,
                 'mismatches': mismatches,
                 'XM': len(mismatches),
                 'XO': len(insertions) + len(deletions),
                }

    return alignment

def extend_perfect_seed(char* query_seq, char* target_seq, int query_start, int query_end, int target_start, int target_end):
    while query_start > 0 and target_start > 0 and (query_seq[query_start - 1] == target_seq[target_start - 1]):
        query_start -= 1
        target_start -= 1

    while query_end < len(query_seq) and target_end < len(target_seq) and (query_seq[query_end] == target_seq[target_end]):
        query_end += 1
        target_end += 1

    return query_start, query_end, target_start

def extend_perfect_seed_with_one_nt_deletion(char* query_seq, char* target_seq, int query_start, int query_end, int target_start, int target_end):
    cdef gained_before = 0
    cdef gained_after = 0
    
    target_start -= 1
    
    while query_start > 0 and target_start > 0 and (query_seq[query_start - 1] == target_seq[target_start - 1]):
        gained_before += 1
        query_start -= 1
        target_start -= 1

    target_end += 1
    
    while query_end < len(query_seq) and target_end < len(target_seq) and (query_seq[query_end] == target_seq[target_end]):
        gained_after += 1
        query_end += 1
        target_end += 1

    return gained_before, gained_after

@cython.boundscheck(False)
def mismatches_at_offset(char* donor, char* target):
    ''' Number of mismatches if donor is aligned to target at each possible offset. '''
    mismatches = np.zeros((len(donor), len(target) - len(donor)), int)
    cdef long [:, ::1] mismatches_view = mismatches
    cdef unsigned int row, col
    
    for col in range(len(target) - len(donor)):
        for row in range(len(donor)):
            if donor[row] != target[col + row]:
                mismatches_view[row, col] += 1
                
    total_mismatches = mismatches.sum(axis=0)
    return total_mismatches

def hamming_distance(char* first, char* second):
    cdef int i
    cdef int d = 0
    cdef int length = len(first)
    
    for i in range(length):
        if first[i] != second[i]:
            d += 1
            
    return d


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hits',
 'hits.parallel',
 'hits.serialize',
 'hits.test',
 'hits.visualize',
 'hits.visualize.interactive']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.4.0,<10.0.0',
 'biopython>=1.81,<2.0',
 'bokeh>=3.0.3,<4.0.0',
 'ipython>=8.10.0,<9.0.0',
 'ipywidgets>=8.0.4,<9.0.0',
 'matplotlib>=3.7.0,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pysam>=0.20.0,<0.21.0',
 'pyyaml>=6.0,<7.0',
 'seaborn>=0.12.2,<0.13.0',
 'statsmodels>=0.13.5,<0.14.0']

setup_kwargs = {
    'name': 'hits-x',
    'version': '0.1.3',
    'description': "Updated and modified from Jeff's version.",
    'long_description': 'utilities for processing high-throughput sequencing experiments (v0.3.3)\n',
    'author': 'hukai916',
    'author_email': '31452631+hukai916@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import numpy as np

extension = 'txt'

def write_file(array, file_name):
    ''' Writes a 1D array into a text file.
    '''
    if array.ndim < 1:
        array = np.atleast_1d(array)
    elif array.ndim == 1:
        np.savetxt(file_name, array, fmt='%i', delimiter='\t')
    else:
        raise RuntimeError('Array has more than 2 dimensions.')

def read_file(file_name):
    ''' Reads the format written by array_file.write into an array.
    '''
    array = np.loadtxt(file_name, dtype=np.int, ndmin=1)
    return array

def combine_data(first_array, second_array):
    ''' Identically shaped arrays are added. Arrays of different shapes are
        zero-padded to the size of the biggest before adding.
    '''
    first_length = len(first_array)
    second_length = len(second_array)
    max_length = max(first_length, second_length)
    combined = np.zeros(max_length, int)
    combined[:first_length] += first_array
    combined[:second_length] += second_array
    return combined



from collections import Counter

extension = 'txt'

def write_file(dict_of_counters, file_name):
    with open(file_name, 'w') as fh:
        for name, counter in sorted(dict_of_counters.items()):
            if '\t' in name:
                raise ValueError('Name of counter can\'t contain a tab')
            fh.write('{0}\n'.format(name))

            for key, count in counter.most_common():
                if isinstance(key, tuple):
                    key = ' '.join(key)
                fh.write('{0}\t{1}\n'.format(key, count))

def read_file(file_name):
    dict_of_counters = {}
    counter = Counter()
    for line in open(file_name):
        line = line.strip()
        if '\t' not in line:
            name = line
            counter = Counter()
            dict_of_counters[name] = counter
        else:
            key, count = line.split('\t')
            counter[key] = int(count)

    return dict_of_counters

def combine_data(first, second):
    for name in second:
        first[name].update(second[name])
    return first



from PIL import ImageColor
import numpy as np

base_to_hsl = {
    'A': 'hsl(120, 100%, {0:d}%)'.format,
    'C': 'hsl(240, 100%, {0:d}%)'.format,
    'G': 'hsl(30, 75%, {0:d}%)'.format,
    'T': 'hsl(0, 100%, {0:d}%)'.format,
}

def _get_base_rgb(base):
    if base == 'N' or base == '-' or base == '.':
        rgb = (0, 0, 0)
    else:
        l = 45
        outline_hsl = base_to_hsl[base](l)
        rgb = ImageColor.getrgb(outline_hsl)
    return rgb

rgbs = {base: _get_base_rgb(base) for base in 'TCAGN-'}
normalized_rgbs = {b: np.array(rgb) / 255. for b, rgb in rgbs.items()}



# cython: language_level=3

from cpython cimport array
cimport cython
import array

cdef int SANGER_OFFSET_typed = 33
cdef unsigned char SANGER_OFFSET_char = 33
SANGER_OFFSET = SANGER_OFFSET_typed

def subtract_sanger_offset(unsigned char[:] q_array):
    cdef unsigned int i, seq_length
    seq_length = len(q_array)
    
    for i in range(seq_length):
        q_array[i] -= SANGER_OFFSET_char

@cython.boundscheck(False)
def process_read(char* seq,
                 char* qual,
                 long[:, ::1] q_array,
                 long[:, ::1] average_q_array,
                 long[:, ::1] c_array,
                 long[:, ::1] c_above_min_q_array,
                 int min_q=0,
                ):
    cdef unsigned int i, q, b, seq_length

    seq_length = len(seq)
    for i in range(seq_length):
        # Automatic type conversion means ord() is unneccesary
        q = qual[i] - SANGER_OFFSET_typed
        q_array[i, q] += 1

        b = seq[i]

        average_q_array[i, b] += q

        c_array[i, b] += 1
        
        if q >= min_q:
            c_above_min_q_array[i, b] += 1

@cython.boundscheck(False)
def process_Alignment(char* seq,
                      char[:] qual,
                      long[:, ::1] q_array,
                      long[:, ::1] average_q_array,
                      long[:, ::1] c_array,
                      long[:, ::1] c_above_min_q_array,
                      int min_q=0,
                     ):
    cdef unsigned int i, q, b, seq_length

    seq_length = len(seq)
    for i in range(seq_length):
        q = qual[i]
        q_array[i, q] += 1

        b = seq[i]
        
        average_q_array[i, b] += q

        c_array[i, b] += 1
        
        if q >= min_q:
            c_above_min_q_array[i, b] += 1

def RSQCI_length(char *quals, int length):
    cdef int i
    for i in range(length):
        if quals[length - 1 - i] - SANGER_OFFSET_typed != 2:
            return i

    return length



import numpy as np

extension = 'npy'

def write_file(type_counts, mismatch_file_name):
    if not str(mismatch_file_name).endswith('.npy'):
        raise ValueError('{0} does not end in .npy'.format(mismatch_file_name))
    np.save(mismatch_file_name, type_counts)

def read_file(mismatch_file_name):
    type_counts = np.load(mismatch_file_name)
    return type_counts

def combine_data(first_type_counts, second_type_counts):
    assert first_type_counts.shape == second_type_counts.shape
    return first_type_counts + second_type_counts



import os
import glob
from collections import defaultdict
from pathlib import Path

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.image
import scipy.signal
import scipy.interpolate
import numpy as np
import ipywidgets
import yaml

from .. import utilities

def load_image(image_fn):
    return matplotlib.image.imread(image_fn)

def load_labels(image_fn):
    labels_fn = '{0}_labels.txt'.format(image_fn[:-len('.tif')])
    labels = [line.strip() for line in open(labels_fn)]
    return labels

def load_annotations(image_fn):
    root, ext = os.path.splitext(image_fn)
    annotation_fn = '{0}.yaml'.format(root)

    if os.path.exists(annotation_fn):
        with open(annotation_fn) as annotation_fh:
            annotations = yaml.load(annotation_fh)
    else:
        annotations = {}

    return annotations

def save_annotations(image_fn, annotations):
    root, ext = os.path.splitext(image_fn)
    annotation_fn = '{0}.yaml'.format(root)

    with open(annotation_fn, 'w') as annotation_fh:
        yaml.dump(annotations, annotation_fh, default_flow_style=False)

def identify_lane_boundaries(image, num_lanes=None, window_size=10):
    ''' Finds lane boundaries by summing image columns, identifying peaks in
    window-summed column sums, then determining widths around peaks at which
    signal has returned to background.
    '''
    cols = image.sum(axis=0)
    window_sums = sum_over_window(cols, window_size)
    peaks, = scipy.signal.argrelmax(window_sums, order=16)

    if num_lanes is not None:
        descending = sorted(peaks, key=lambda p: window_sums[p], reverse=True)
        peaks = sorted(descending[:num_lanes])

    mins_between = find_mins_between(cols, peaks)
    half_widths = []
    for p, (l, r) in zip(peaks, mins_between):
        right_threshold = cols[r] + 0.05 * (cols[p] - cols[r])
        # argmax gives index of first true value
        right_width = np.argmax(cols[p:r] <= right_threshold)
        
        left_threshold = cols[l] + 0.05 * (cols[p] - cols[l])
        left_width = np.argmax(cols[l:p + 1][::-1] <= left_threshold)
    
        half_width = max(right_width, left_width)
        half_widths.append(half_width)

    half_width = np.mean(half_widths) 
    either_side = int(half_width)
    boundaries = [(p - either_side, p + either_side) for p in peaks]
    return boundaries

def extract_profiles(image, boundaries):
    profiles = []
    for start, end in boundaries:
        start = max(start, 0)
        lane = image[:, start:end + 1]
        rows, cols = lane.shape
        raw = lane.sum(axis=1)[::-1]
        normalized = np.true_divide(raw, cols)
        profiles.append(normalized)
    return profiles

def extract_background(image, boundaries):
    profiles = []
    for (_, end), (start, _) in zip(boundaries[:-1], boundaries[1:]):
        lane = image[:, end + 1 + 1:start + 1 - 1]
        rows, cols = lane.shape
        raw = lane.sum(axis=1)[::-1]
        normalized = np.true_divide(raw, cols)
        profiles.append(normalized)
    return profiles

def sum_over_window(array, either_side):
    summed = np.zeros(len(array))
    cumulative = np.concatenate(([0], np.cumsum(array)))
    window = 2 * either_side + 1
    summed[either_side:-either_side] = cumulative[window:] - cumulative[:-window]
    return summed

def get_ladder100_peaks(ys):
    ys = np.asarray(ys)
    lengths = [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1200, 1517]
    
    relative_maxes, = scipy.signal.argrelmax(ys, order=4)

    # Filter out peaks that aren't sufficiently high.
    median = np.percentile(ys, 50)
    overall_max = np.max(ys)
    threshold = median + (overall_max - median) * 0.3
    high_enough = relative_maxes[ys[relative_maxes] > threshold] 

    # Filter out peaks that are too close to the top or bottom.
    candidates = [c for c in high_enough if 0.05 * len(ys) < c < 0.95 * len(ys)]

    if len(candidates) == len(lengths):
        peaks = dict(zip(lengths, candidates))
    else:
        # Peaks are probably too close together to call.
        # Give up on the higher ones.
        peaks = dict(zip(lengths[:5], candidates[:5]))
        peaks[lengths[-1]] = candidates[-1]
    
        relative_maxes,  = scipy.signal.argrelmax(ys, order=10)
        between = [x for x in relative_maxes if peaks[500] < x < peaks[1517]]
        if len(between) > 0:
            peaks[1000] = max(between, key=lambda x: ys[x])
        
    return peaks

def find_mins_between(ys, peaks):
    ''' For each index in peaks, returns the indices of the minimum values
    between peaks on either side.
    '''
    either_side = []
    for i, p in enumerate(peaks):
        if i == 0:
            left = 0
        else:
            left = peaks[i - 1]

        if i == len(peaks) - 1:
            right = len(ys)
        else:
            right = peaks[i + 1]

        min_left = left + np.argmin(ys[left:p])
        min_right = p + np.argmin(ys[p:right])
        either_side.append((min_left, min_right))

    return either_side

def compute_peak_to_troughs(ys, peaks):
    ''' For each index in peaks, returns the peak-to-trough ratio of
    the value at that peak to the average of the minimum values between
    the peak and its adjacent peaks.
    '''
    mins_between = find_mins_between(ys, peaks)
    trough_means = [np.mean([ys[left], ys[right]]) for left, right in mins_between]
    ratios = [ys[p] / trough_mean for p, trough_mean in zip(peaks, trough_means)]
    return ratios

def get_ladder10_peaks(ys, ladder100_peaks=None):
    # Strategy: identify the 100 bp beak by finding the relative max closest
    # to the 100 bp peak in the 100bp ladder.
    # March up and down relative maxes from there to assign 10 bp peaks.
    # To avoid false positive peaks between the lower ladder components,
    # require them to be the max in a larger index winow (via the order
    # argument to argrelmax).
    
    ys = np.asarray(ys)
    peaks = {}
    
    def get_relative_max_xs(order):
        xs, = scipy.signal.argrelmax(ys, order=order)
        return xs

    big_peak_xs = get_relative_max_xs(50)

    if ladder100_peaks is not None:
        peaks[100] = min(big_peak_xs, key=lambda x: abs(ladder100_peaks[100] - x))
    else:
        peaks[100] = sorted(big_peak_xs, key=lambda x: ys[x])[-2]
    
    # below 100
    xs = get_relative_max_xs(10)
    peak_to_troughs = compute_peak_to_troughs(ys, xs)
    below_100 = [x for x, r in zip(xs, peak_to_troughs)
                 if x < peaks[100] and r > 1.1
                ]
    
    for length, x in zip(np.arange(90, 10, -10), below_100[::-1]):
        peaks[length] = x
    
    # above 100
    xs = get_relative_max_xs(3)
    above_100 = [x for x in xs if x > peaks[100]]
    
    for length, x in zip(np.arange(110, 160, 10), above_100):
        peaks[length] = x
        
    return peaks

def make_conversion_functions(lanes):
    ladder100_peaks = get_ladder100_peaks(lanes['ladder100'])
    ladder10_peaks = get_ladder10_peaks(lanes['ladder10'], ladder100_peaks)
    for l in [100, 110, 120, 130, 140, 150]:
        ladder10_peaks.pop(l, None)
    length_to_x = dict(ladder10_peaks.items() + ladder100_peaks.items())
    lengths = sorted(length_to_x)
    xs = [length_to_x[l] for l in lengths]

    log_length_to_x = scipy.interpolate.interp1d(np.log10(lengths), xs, 'linear', fill_value='extrapolate')
    length_to_x = lambda l: log_length_to_x(np.log10(l))
    
    x_to_log_length = scipy.interpolate.interp1d(xs, np.log10(lengths), 'linear', fill_value='extrapolate')
    x_to_length = lambda x: 10**x_to_log_length(x)

    return length_to_x, x_to_length

def analyze_image(image, annotations, window_size):
    def contain_overlap(boundaries):
        overlapping = False
        for (_, end), (start, _) in utilities.pairwise(boundaries):
            if end >= start:
                overlapping = True
        return overlapping

    def invalid(boundaries, labels):
        wrong_number = labels is not None and len(labels) != len(boundaries)
        return wrong_number or contain_overlap(boundaries)

    labels = annotations.get('labels')
    if labels is None:
        num_lanes = None
    else:
        num_lanes = len(labels)

    boundaries = identify_lane_boundaries(image, num_lanes=num_lanes, window_size=window_size)
    profiles = extract_profiles(image, boundaries)
    
    if labels is None:
        labels = range(1, len(boundaries) + 1)
    labels = [str(l) for l in labels]

    #if invalid(boundaries, labels):
    #    labels = map(str, range(1, len(boundaries) + 1))

    lanes = dict(zip(labels, profiles))

    return labels, boundaries, profiles, lanes

def plot_gel(image_fn,
             vertical_range=(0, 1),
             highlight=None,
             contrast=1.0,
             show_expected='both',
             invert=False,
             window_size=10,
            ):
    image = load_image(image_fn)
    annotations = load_annotations(image_fn)

    labels, boundaries, profiles, lanes = analyze_image(image, annotations, window_size)

    fig_height = 8.
    plot_width = 12.
    rows, cols = image.shape
    im_width = fig_height * cols / rows

    gridspec_kw = dict(width_ratios=[im_width, plot_width], wspace=0)
    
    fig, (im_ax, line_ax) = plt.subplots(1, 2,
                                         figsize=(plot_width + im_width, fig_height),
                                         gridspec_kw=gridspec_kw,
                                        )

    kwargs = {
        'line': {
            'highlight': dict(alpha=0.95, linewidth=1),
            'nonhighlight': dict(alpha=0.1, linewidth=0.5),
            'uniform': dict(alpha=0.9, linewidth=0.7),
        },
        'boundary': {
            'highlight': dict(alpha=0.95, linewidth=1.2),
            'nonhighlight': dict(alpha=0.2, linewidth=1),
            'uniform': dict(alpha=0.5, linewidth=1),
        },
        'text': {
            'highlight': dict(alpha=0.95),
            'nonhighlight': dict(alpha=0.1),
            'uniform': dict(alpha=0.9),
        },
    }
    
    label_to_kwargs = defaultdict(dict)

    if highlight is None:
        highlight = []
    
    for i, label in enumerate(labels):
        color = 'C{0}'.format(i % 10)
        
        if len(highlight) > 0:
            if label in highlight:
                key = 'highlight'
            else:
                key = 'nonhighlight'
        else:
            key = 'uniform'
        
        for kind in kwargs:
            copy = kwargs[kind][key].copy()
            copy['color'] = color
            label_to_kwargs[kind][label] = copy
    
    max_y = 0
    min_y = 1e99

    for i, label in enumerate(labels):
        if 'ladder' in label and not label in highlight:
            continue

        ys = lanes[label]
        xs = np.arange(len(ys))

        line_ax.plot(xs, ys, 'o-', label=label, markersize=0.3, **label_to_kwargs['line'][label])

        max_y = max(max_y, max(ys))
        min_y = min(min_y, min(ys))

    peaks = None
    if 'ladder100' in lanes:
        ladder100_peaks = get_ladder100_peaks(lanes['ladder100'])

        if 'ladder10' in lanes:
            ladder10_peaks = get_ladder10_peaks(lanes['ladder10'], ladder100_peaks)
    
            # Only include the 100 bp peak from the 10 bp ladder.
            ladder100_peaks.pop(100)

            peaks = list(ladder100_peaks.items()) + list(ladder10_peaks.items())
        else:
            peaks = list(ladder100_peaks.items())

    elif 'ladder10' in lanes:
        ladder10_peaks = get_ladder10_peaks(lanes['ladder10'])
        peaks = list(ladder10_peaks.items())
    
    if peaks is not None:
        major_peaks = [
            100,
            200,
            500,
            1000,
            1517,
        ]

        for length, x in peaks:
            alpha = 0.3 if length in major_peaks else 0.05
            line_ax.axvline(x, color='black', alpha=alpha)

        line_ax.set_xticks([x for length, x in peaks])
        line_ax.set_xticklabels([str(length) for length, x in peaks], rotation=-90, ha='center', size=8)

        if show_expected:
            length_to_x, x_to_length = make_conversion_functions(lanes)
            expected = annotations.get('expected', {})
            for length, name in expected.items():
                x = length_to_x(length)
                if show_expected in {'both', 'plot'}:
                    line_ax.axvline(x, color='black', linestyle='--')
                    line_ax.annotate(name,
                                     xy=(x, 1),
                                     xycoords=('data', 'axes fraction'),
                                     xytext=(0, 2),
                                     textcoords='offset points',
                                     va='bottom',
                                     ha='center',
                                    )

                if invert:
                    line_color = 'black'
                else:
                    line_color = 'white'
                y = rows - x

                if show_expected in {'both', 'image'}:
                    im_ax.plot([0.99 * cols, cols], [y, y], color=line_color)
                    im_ax.plot([0, 0.01 * cols], [y, y], color=line_color)

                    im_ax.annotate(name,
                                   xy=(1, y),
                                   xycoords=('axes fraction', 'data'),
                                   xytext=(2, 0),
                                   textcoords='offset points',
                                   fontsize=8,
                                   va='center',
                                   ha='left',
                                  )

    else:
        line_ax.set_xticks([])

    line_ax.set_yticks([])
    
    
    if invert:
        im_cmap = matplotlib.cm.binary
    else:
        im_cmap = matplotlib.cm.binary_r

    image_range = image.max() - image.min()
    im_ax.imshow(image,
                 cmap=im_cmap,
                 vmin=image.min(),
                 vmax=image.min() + image_range * contrast,
                )
    im_ax.set_xticks([])
    im_ax.set_yticks([])
    
    for i, (label, (start, end)) in enumerate(zip(labels, boundaries)):
        im_ax.axvline(start, **label_to_kwargs['boundary'][label])
        im_ax.axvline(end, **label_to_kwargs['boundary'][label])
        im_ax.annotate(label,
                       xy=(np.mean((start, end)), 1),
                       xytext=(0, 2),
                       xycoords=('data', 'axes fraction'),
                       textcoords='offset points',
                       rotation=45,
                       fontsize=10,
                       va='bottom',
                       **label_to_kwargs['text'][label])
        
    #legend = line_ax.legend(loc='upper left', framealpha=0.5)
    #for text in legend.get_texts():
    #    label = text.get_text()
    #    text.set(**label_to_kwargs['text'][label])
   
    x_min, x_max = map(int, np.asarray(vertical_range) * len(xs))
    line_ax.set_xlim(x_min, x_max)
    y_range = max_y - min_y
    line_ax.set_ylim(min_y - y_range * 0.02, max_y + y_range * 0.02)
    
    im_ax.autoscale(False)

    if vertical_range != (0, 1):
        y_min, y_max = vertical_range

        if y_min == 0:
            y_min = 0.002
        if y_max == 1:
            y_max= 0.999

        if invert:
            line_color = 'black'
        else:
            line_color = 'white'

        line_kwargs = dict(transform=im_ax.transAxes, color=line_color)
        im_ax.plot([0.005, 0.005, 0.05], [y_min + 0.05, y_min, y_min], **line_kwargs)
        im_ax.plot([0.005, 0.005, 0.05], [y_max - 0.05, y_max, y_max], **line_kwargs)
        im_ax.plot([1 - 0.005, 1 - 0.005, 1 - 0.05], [y_min + 0.05, y_min, y_min], **line_kwargs)
        im_ax.plot([1 - 0.005, 1 - 0.005, 1 - 0.05], [y_max - 0.05, y_max, y_max], **line_kwargs)
        
    head, tail = os.path.split(image_fn)
    line_ax.set_title(tail, y=1.05)
    
    return fig

def plot_gels_interactive(image_fns, **kwargs):

    def make_tab(image_fn):
        def generate_figure(highlight, vertical_range, contrast, invert, show_expected, window_size, notes):
            fig = plot_gel(image_fn,
                           highlight=highlight,
                           vertical_range=vertical_range,
                           contrast=contrast,
                           invert=invert,
                           show_expected=show_expected,
                           window_size=window_size,
                           **kwargs)
            fig.axes[1].annotate(notes,
                                 xy=(0, 1),
                                 xycoords='axes fraction',
                                 xytext=(5, 2),
                                 textcoords='offset points',
                                 ha='left',
                                 va='bottom',
                                 size=8,
                                )
            plt.show()
            return fig

        annotations = load_annotations(image_fn)
        labels = annotations.get('labels', [])
    
        default_filename = str(Path.home() / Path(image_fn).with_suffix('.png').name)

        widgets = {
            'vertical_range': ipywidgets.FloatRangeSlider(
                value=[0, 1],
                continuous_update=False,
                min=0,
                max=1,
                step=0.01,
                layout=ipywidgets.Layout(height='200px'),
                description='Vertical range:',
                style={'description_width': 'initial'},
                orientation='vertical',
            ),
            'window_size': ipywidgets.IntSlider(
                value=10,
                continuous_update=False,
                min=1,
                max=30,
                step=1,
                layout=ipywidgets.Layout(height='200px'),
                description='Window size:',
                style={'description_width': 'initial'},
                orientation='vertical',
            ),
            'highlight': ipywidgets.SelectMultiple(
                options=labels,
                value=[],
                layout=ipywidgets.Layout(height='200px'),
                description='Highlight:',
            ),
            'contrast': ipywidgets.FloatSlider(
                value=1.0,
                continuous_update=False,
                min=0,
                max=1,
                step=0.05,
                layout=ipywidgets.Layout(height='200px'),
                description='Contrast:',
                style={'description_width': 'initial'},
                orientation='vertical',
            ),
            'invert': ipywidgets.ToggleButton(
                value=False,
                description='Invert',
                icon='check',
            ),
            'show_expected': ipywidgets.Dropdown(
                options={'neither': False,
                         'plot': 'plot',
                         'image': 'image',
                         'both': 'both',
                        },
                value=False,
                description='Show expected',
            ),
            'save': ipywidgets.Button(
                description='Save snapshot',
            ),
            'file_name': ipywidgets.Text(
                value=default_filename,
            ),
            'close': ipywidgets.Button(
                description='Close tab',
            ),
            'notes': ipywidgets.Textarea(
                description='Notes:',
                layout=ipywidgets.Layout(height='200px', width='400px'),
            ),
            'labels': ipywidgets.Textarea(
                description='Lane labels:',
                layout=ipywidgets.Layout(height='200px', width='400px'),
            ),
            'expected': ipywidgets.Textarea(
                description='Expected:',
                layout=ipywidgets.Layout(height='200px', width='400px'),
            ),
            'update': ipywidgets.Button(
                description='Update',
            ),
        }

        def save(_):
            fig = interactive.result
            fn = widgets['file_name'].value
            fig.savefig(fn, bbox_inches='tight')

        widgets['save'].on_click(save)
        
        def close(_):
            titles = [tabs.get_title(i) for i, c in enumerate(tabs.children) if i != tabs.selected_index]
            tabs.children = [c for i, c in enumerate(tabs.children) if i != tabs.selected_index]
            for i, title in enumerate(titles):
                tabs.set_title(i, title)

        widgets['close'].on_click(close)

        interactive = ipywidgets.interactive(generate_figure, **widgets)
        output = interactive.children[-1]
        interactive.update()

        widgets['invert'].value = annotations.get('invert', False)
        widgets['contrast'].value = annotations.get('contrast', 1.0)
        widgets['window_size'].value = annotations.get('window_size', 10)
        widgets['notes'].value = annotations.get('notes', '')
        widgets['labels'].value = '\n'.join(annotations.get('labels', []))
        expected = annotations.get('expected', {})
        lines = ['{0}: {1}'.format(length, name) for length, name in sorted(expected.items())]
        widgets['expected'].value = '\n'.join(lines)

        def update_annotations(_):
            def get_lines(key):
                lines = [str(v) for v in widgets[key].value.split('\n')]
                if lines == ['']:
                    lines = []
                return lines

            annotations['notes'] = str(widgets['notes'].value)

            labels = get_lines('labels')
            widgets['highlight'].options = labels

            expected = {}
            for line in get_lines('expected'):
                length, name = line.split(': ')
                length = int(length)
                expected[length] = name
            
            annotations['expected'] = expected
            annotations['labels'] = labels
            annotations['invert'] = widgets['invert'].value
            annotations['contrast'] = widgets['contrast'].value
            annotations['window_size'] = widgets['window_size'].value

            save_annotations(image_fn, annotations)
            interactive.update()

        widgets['update'].on_click(update_annotations)

        def group_widgets(keys, kind):
            if kind == 'row':
                Box = ipywidgets.HBox
            elif kind == 'col':
                Box = ipywidgets.VBox

            keys = [widgets.get(k, k) for k in keys]

            return Box(keys)

        make_row = lambda keys: group_widgets(keys, 'row')
        make_col = lambda keys: group_widgets(keys, 'col')

        layout = make_col([
            make_row([output]),
            make_row(['highlight',
                      'vertical_range',
                      'contrast',
                      'window_size',
                      make_col([
                          'invert',
                      ]),
                      make_col([
                          'save',
                          'close',
                          'update',
                      ]),
                      make_col([
                          'file_name',
                          'show_expected',
                      ]),
                     ]),
            make_row(['labels',
                      'expected',
                      'notes',
                     ]),
        ])

        return layout 

    shorten_fn = lambda fn: os.path.split(fn)[1][:-len('_cropped.tif')]

    tabs = ipywidgets.Tab()
    tabs.children = [make_tab(image_fn) for image_fn in image_fns]
    for i, image_fn in enumerate(image_fns):
        shortened = shorten_fn(image_fn)
        tabs.set_title(i, shortened)
            
    fns = sorted(glob.glob('/home/jah/cropped/*.tif'))
    file_names = ipywidgets.Select(
        options=[(shorten_fn(fn), fn) for fn in fns],
        value=None,
        layout=ipywidgets.Layout(width='300px', height='800px'),
    )

    def open_file(change):
        existing_titles = [tabs.get_title(i) for i in range(len(tabs.children))]
        image_fn = change['new']
        shortened = shorten_fn(image_fn)
        if shortened not in existing_titles:
            new_tab = make_tab(image_fn)
            tabs.children = tabs.children + (new_tab,)
            tabs.set_title(len(tabs.children) - 1, shortened)
        
        # Make the selected file the active tab
        titles = [tabs.get_title(i) for i in range(len(tabs.children))]
        tabs.selected_index = titles.index(shortened)

    def sync_menu_selection(change):
        if change['new'] is None:
            # All tabs are closed.
            file_names.value = None
        else:
            shorteneds = [s for s, f in file_names.options]
            fulls = [f for s, f in file_names.options]
            new_title = tabs.get_title(change['new'])
            index = shorteneds.index(new_title)
            file_names.value = fulls[index]

    tabs.observe(sync_menu_selection, names='selected_index')

    file_names.observe(open_file, names='value')

    file_col = ipywidgets.VBox([file_names])

    return ipywidgets.HBox([file_col, tabs])
